(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright (c) 2013-2014 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: scalar.ml,v 1.34 2014/06/10 15:08:35 fclement Exp $ *)

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Structural_mapping = Umatrix_structural_mapping;;

module Config = Configuration;;

let fun_name = Printf.sprintf "Scalar.%s";;

(* Not imposing an explicit formatter here cause an error with output_value
   with message: "abstract value (Custom)". *)
let s_print_debug fname ident print =
  Config.print_debug (fun_name fname) ident print Format.err_formatter
;;

open Refinement_types;;
open Refinement_types_print;;
open Parameter;;
open Zonation;;

module Make =
  functor (Of : Uordered_field_spec.ORDERED_FIELD) ->
struct
  include Of;;
  
  let sum = function
    | { Mapping.domain = dom; images = Mapping.Constant y; } ->
        of_nat (Set.cardinal dom) *. y
    | { Mapping.images = Mapping.Images ims; _; } ->
        Array.fold_left ( +. ) _0_ ims
  ;;
  
  let first_order_indicator g_zone (zt, zf) =
    let indict = sum (Mapping.restriction g_zone zt) in
    let indicf = sum (Mapping.restriction g_zone zf) in
    abs (indict -. indicf)
  ;;
  
(*
  let value_is_zero x = is_zero (of_float (Config.get_minimum_gradient ())) x;;
  let vector_is_zero = Uarray.for_all value_is_zero;;
  
  let gradient_is_zero = function
    | { Mapping.images = Mapping.Constant gy; _; } -> value_is_zero gy
    | { Mapping.images = Mapping.Images gs; _; } -> vector_is_zero gs
  ;;
*)
  
  let make_overall_best_dicho g_ss dom =
    s_print_debug "make_overall_best_dicho" "g_ss"
      (print_gradient print) g_ss;
    let p idx _uidx = Mapping.image g_ss idx < _0_ in
    Set.dichotomy p dom
  ;;
  
  let make_best_in_a_family_dichos nct bcfl1 dom =
    let bds = bounds_of_domain dom in
    Ulist.concat
      (Ulist1.to_list
         (Ulist1.rev_map
            (dichotomies_from_family bds dom nct) bcfl1))
  ;;
  
  module Key = struct type t = Mapping.domain;; end;;
  module Hashtbl = Uhashtbl_eq.Make (Key);;
  
  let memo_cuttings_tbl = Hashtbl.create 1001;;
  
  let memo_make_best_in_a_family_dichos nct bcfl1 dom =
    if Hashtbl.mem memo_cuttings_tbl dom then
      Hashtbl.find memo_cuttings_tbl dom else
    let dichos = make_best_in_a_family_dichos nct bcfl1 dom in
    Hashtbl.add memo_cuttings_tbl dom dichos;
    dichos
  ;;
  
  let get_dicho g_ss dom =
    match make_overall_best_dicho g_ss dom with
    | None -> []
    | Some dicho -> [ dicho; ]
  ;;
  
  (* Fixme: argument g of type 'a gradient should be replaced by an 'a
     partition computed using Zonation.partition_of_gradient.
     This means data structure Structural_mapping.t should be modified. *)
  let cut dim dm cs g _zn { Mapping.domain = dom; _; } =
    let ss = Set.subset_of_full g.Mapping.domain dom in
    let g_ss = Mapping.restriction g ss in
    let dichos =
      (*if gradient_is_zero g_ss then [] else*)
      match (dm, cs) with
      | (Vector, Overall_best) ->
          if dim <> Unatural._1_ then [] else
          get_dicho g_ss dom
          (* Fixme: We have to use the infinite-norm here
             (and extend it to Best_component_only). *)
      | (Best_component_only, Overall_best) ->
          get_dicho g_ss dom
      | (Vector, Best_in_a_family (nct, bcfl1)) ->
          memo_make_best_in_a_family_dichos nct bcfl1 dom
      | (Best_component_only, Best_in_a_family (nct, bcfl1)) ->
          make_best_in_a_family_dichos nct bcfl1 dom in
    Ulist.map
      (fun dicho ->
        let indic = first_order_indicator g_ss dicho in
        { cutting = dicho; first_order_indicator = indic; })
      dichos
  ;;
  
  let initial_zonation pp1 =
    let dim = Config.get_dimension ()
    and dm = Config.get_dimension_management ()
    and cs = Config.get_cutting_strategy () in
    of_partition (cut dim dm cs) pp1
  ;;
end
;;

module Float = Make (Uordered_field.Float);;
